import socket
import subprocess


def get_free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def is_rocm_system() -> bool:
    try:
        subprocess.check_output(["rocm-smi"])
        return True
    except FileNotFoundError:
        return False


def is_nvidia_system() -> bool:
    try:
        subprocess.check_output(["nvidia-smi"])
        return True
    except FileNotFoundError:
        return False



import asyncio
from dataclasses import dataclass
from logging import getLogger
from typing import List, Literal, Optional, Union

import numpy as np

from .inference_server import InferenceServer, InferenceServerConfig
from .utils import is_nvidia_system

LOGGER = getLogger("Text-Embedding-Inference")


Pooling_Literal = Literal["cls", "mean"]
DType_Literal = Literal["float32", "float16"]


@dataclass
class TEIConfig(InferenceServerConfig):
    # Docker options
    image: str = "ghcr.io/huggingface/text-embeddings-inference:cpu-latest"
    # Launcher options
    model_id: str = "bert-base-uncased"
    revision: str = "main"
    dtype: Optional[DType_Literal] = None
    pooling: Optional[Pooling_Literal] = None
    # Concurrency options
    max_concurrent_requests: int = 512

    def __post_init__(self) -> None:
        super().__post_init__()

        if is_nvidia_system() and "cpu" in self.image:
            LOGGER.warning(
                "Your system has NVIDIA GPU, but you are using a CPU image."
                "Consider using a GPU image for better performance."
            )


class TEI(InferenceServer):
    NAME: str = "Text-Embedding-Inference"
    SUCCESS_SENTINEL: str = "Ready"
    FAILURE_SENTINEL: str = "Error"

    def __init__(self, config: TEIConfig) -> None:
        super().__init__(config)

    async def single_client_call(self, text: str, **kwargs) -> np.ndarray:
        async with self.semaphore:
            output = await self.client.feature_extraction(text=text, **kwargs)
            return output

    async def batch_client_call(self, text: List[str], **kwargs) -> List[np.ndarray]:
        output = await asyncio.gather(*[self.single_client_call(t, **kwargs) for t in text])
        return output

    def encode(self, text: Union[str, List[str]], **kwargs) -> Union[np.ndarray, List[np.ndarray]]:
        if isinstance(text, str):
            output = asyncio.run(self.single_client_call(text, **kwargs))
            return output
        elif isinstance(text, list):
            output = asyncio.run(self.batch_client_call(text, **kwargs))
            return output
        else:
            raise ValueError(f"Unsupported input type: {type(text)}")



import asyncio
from dataclasses import dataclass
from logging import getLogger
from typing import Literal, Optional, Union

from .inference_server import InferenceServer, InferenceServerConfig
from .utils import is_rocm_system

LOGGER = getLogger("Text-Generation-Inference")

Shareded_Literal = Literal["true", "false"]
DType_Literal = Literal["float32", "float16", "bfloat16"]
Quantize_Literal = Literal["bitsandbytes-nf4", "bitsandbytes-fp4", "gptq"]


@dataclass
class TGIConfig(InferenceServerConfig):
    # Docker options
    image: str = "ghcr.io/huggingface/text-generation-inference:latest"
    # Launcher options
    model_id: str = "gpt2"
    revision: str = "main"
    num_shard: Optional[int] = None
    dtype: Optional[DType_Literal] = None
    enable_cuda_graphs: Optional[bool] = None
    sharded: Optional[Shareded_Literal] = None
    quantize: Optional[Quantize_Literal] = None
    disable_custom_kernels: Optional[bool] = None
    trust_remote_code: Optional[bool] = None
    # Concurrency options
    max_concurrent_requests: int = 128

    def __post_init__(self) -> None:
        super().__post_init__()

        if is_rocm_system() and "rocm" not in self.image:
            LOGGER.warning(
                "You are running on a ROCm system but the image is not rocm specific. "
                "Add 'rocm' to the image name to use the rocm specific image."
            )
            self.image += "-rocm"


class TGI(InferenceServer):
    NAME: str = "Text-Generation-Inference"
    SUCCESS_SENTINEL: str = "Connected"
    FAILURE_SENTINEL: str = "Error"

    def __init__(self, config: TGIConfig) -> None:
        super().__init__(config)

    async def single_client_call(self, prompt: str, **kwargs) -> str:
        async with self.semaphore:
            output = await self.client.text_generation(prompt=prompt, **kwargs)
            return output

    async def batch_client_call(self, prompt: list, **kwargs) -> list:
        output = await asyncio.gather(*[self.single_client_call(prompt=p, **kwargs) for p in prompt])
        return output

    def generate(self, prompt: Union[str, list], **kwargs) -> Union[str, list]:
        if isinstance(prompt, str):
            output = asyncio.run(self.single_client_call(prompt, **kwargs))
            return output
        elif isinstance(prompt, list):
            output = asyncio.run(self.batch_client_call(prompt, **kwargs))
            return output
        else:
            raise ValueError(f"Unsupported input type: {type(prompt)}")



from pathlib import Path

from setuptools import find_packages, setup

PY_TXI_VERSION = "0.5.1"

common_setup_kwargs = {
    "author": "Ilyas Moutawwakil",
    "author_email": "ilyas.moutawwakil@gmail.com",
    "description": "A Python wrapper around TGI and TEI servers",
    "keywords": ["tgi", "llm", "tei", "embedding", "huggingface", "docker", "python"],
    "url": "https://github.com/IlyasMoutawwakil/py-txi",
    "long_description_content_type": "text/markdown",
    "long_description": (Path(__file__).parent / "README.md").read_text(encoding="UTF-8"),
    "platforms": ["linux", "windows", "macos"],
    "classifiers": [
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
    ],
}


setup(
    name="py-txi",
    version=PY_TXI_VERSION,
    packages=find_packages(),
    install_requires=["docker", "huggingface-hub", "numpy", "aiohttp"],
    extras_require={"quality": ["ruff"], "testing": ["pytest"]},
    **common_setup_kwargs,
)



import asyncio
import os
import re
import time
from abc import ABC
from dataclasses import asdict, dataclass, field
from logging import INFO, basicConfig, getLogger
from typing import Any, Dict, List, Optional, Union

import docker
import docker.errors
import docker.types
from huggingface_hub import AsyncInferenceClient

from .utils import get_free_port

basicConfig(level=INFO)

DOCKER = docker.from_env()
LOGGER = getLogger("Inference-Server")


@dataclass
class InferenceServerConfig:
    # Image to use for the container
    image: str
    # Shared memory size for the container
    shm_size: str = "1g"
    # List of custom devices to forward to the container e.g. ["/dev/kfd", "/dev/dri"] for ROCm
    devices: Optional[List[str]] = None
    # NVIDIA-docker GPU device options e.g. "all" (all) or "0,1,2,3" (ids) or 4 (count)
    gpus: Optional[Union[str, int]] = None

    ports: Dict[str, Any] = field(
        default_factory=lambda: {"80/tcp": ("127.0.0.1", 0)},
        metadata={"help": "Dictionary of ports to expose from the container."},
    )
    volumes: Dict[str, Any] = field(
        default_factory=lambda: {os.path.expanduser("~/.cache/huggingface/hub"): {"bind": "/data", "mode": "rw"}},
        metadata={"help": "Dictionary of volumes to mount inside the container."},
    )
    environment: Dict[str, str] = field(
        default_factory=lambda: {"HUGGINGFACE_HUB_TOKEN": os.environ.get("HUGGINGFACE_HUB_TOKEN", "")},
        metadata={"help": "Dictionary of environment variables to forward to the container."},
    )

    max_concurrent_requests: Optional[int] = None
    timeout: int = 60

    def __post_init__(self) -> None:
        if self.ports["80/tcp"][1] == 0:
            LOGGER.info("\t+ Getting a free port for the server")
            self.ports["80/tcp"] = (self.ports["80/tcp"][0], get_free_port())


class InferenceServer(ABC):
    NAME: str = "Inference-Server"
    SUCCESS_SENTINEL: str = "Success"
    FAILURE_SENTINEL: str = "Failure"

    def __init__(self, config: InferenceServerConfig) -> None:
        self.config = config

        try:
            LOGGER.info(f"\t+ Checking if {self.NAME} image is available locally")
            DOCKER.images.get(self.config.image)
            LOGGER.info(f"\t+ {self.NAME} image found locally")
        except docker.errors.ImageNotFound:
            LOGGER.info(f"\t+ {self.NAME} image not found locally, pulling from Docker Hub")
            DOCKER.images.pull(self.config.image)

        if self.config.gpus is not None and isinstance(self.config.gpus, str) and self.config.gpus == "all":
            LOGGER.info("\t+ Using all GPU(s)")
            self.device_requests = [docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])]
        elif self.config.gpus is not None and isinstance(self.config.gpus, int):
            LOGGER.info(f"\t+ Using {self.config.gpus} GPU(s)")
            self.device_requests = [docker.types.DeviceRequest(count=self.config.gpus, capabilities=[["gpu"]])]
        elif (
            self.config.gpus is not None
            and isinstance(self.config.gpus, str)
            and re.match(r"^\d+(,\d+)*$", self.config.gpus)
        ):
            LOGGER.info(f"\t+ Using GPU(s) {self.config.gpus}")
            self.device_requests = [docker.types.DeviceRequest(device_ids=[self.config.gpus], capabilities=[["gpu"]])]
        else:
            LOGGER.info("\t+ Not using any GPU(s)")
            self.device_requests = None

        LOGGER.info(f"\t+ Building {self.NAME} command")
        self.command = []
        for k, v in asdict(self.config).items():
            if k in InferenceServerConfig.__annotations__:
                continue
            elif v is not None:
                if isinstance(v, bool):
                    self.command.append(f"--{k.replace('_', '-')}")
                else:
                    self.command.append(f"--{k.replace('_', '-')}={str(v).lower()}")

        address, port = self.config.ports["80/tcp"]
        self.url = f"http://{address}:{port}"

        LOGGER.info(f"\t+ Running {self.NAME} container")
        self.container = DOCKER.containers.run(
            image=self.config.image,
            ports=self.config.ports,
            volumes=self.config.volumes,
            devices=self.config.devices,
            shm_size=self.config.shm_size,
            environment=self.config.environment,
            device_requests=self.device_requests,
            command=self.command,
            auto_remove=True,
            detach=True,
        )

        LOGGER.info(f"\t+ Streaming {self.NAME} server logs")
        for line in self.container.logs(stream=True):
            log = line.decode("utf-8").strip()
            if self.SUCCESS_SENTINEL.lower() in log.lower():
                LOGGER.info(f"\t {log}")
                break
            elif self.FAILURE_SENTINEL.lower() in log.lower():
                LOGGER.info(f"\t {log}")
                raise Exception(f"{self.NAME} server failed to start")
            else:
                LOGGER.info(f"\t {log}")

        try:
            asyncio.set_event_loop(asyncio.get_event_loop())
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())

        self.semaphore = asyncio.Semaphore(self.config.max_concurrent_requests)

        LOGGER.info(f"\t+ Waiting for {self.NAME} server to be ready")
        start_time = time.time()
        while time.time() - start_time < self.config.timeout:
            try:
                if not hasattr(self, "client"):
                    self.client = AsyncInferenceClient(model=self.url)

                asyncio.run(self.single_client_call(f"Hello {self.NAME}!"))
                LOGGER.info(f"\t+ Connected to {self.NAME} server successfully")
                break
            except Exception:
                LOGGER.info(f"\t+ {self.NAME} server is not ready yet, waiting 1 second")
                time.sleep(1)

    async def single_client_call(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    async def batch_client_call(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def close(self) -> None:
        if hasattr(self, "container"):
            LOGGER.info("\t+ Stoping Docker container")
            self.container.stop()
            self.container.wait()
            LOGGER.info("\t+ Docker container stopped")
            del self.container

        if hasattr(self, "semaphore"):
            self.semaphore
            del self.semaphore

        if hasattr(self, "client"):
            del self.client

    def __del__(self) -> None:
        self.close()
